package cn.iocoder.yudao.module.member.api.score;

import cn.iocoder.yudao.framework.apollo.core.event.export.MemberScoreDetailExpireEvent;
import cn.iocoder.yudao.module.member.api.score.dto.*;
import cn.iocoder.yudao.module.member.dal.dataobject.memberUserScore.MemberUserScoreDO;
import cn.iocoder.yudao.module.member.dal.dataobject.memberUserScoreDetail.MemberUserScoreDetailDO;
import cn.iocoder.yudao.module.member.dal.dataobject.memberUserScoreDetailReleation.MemberUserScoreDetailReleationDO;
import cn.iocoder.yudao.module.member.dto.ScoreDetailChangeDto;
import cn.iocoder.yudao.module.member.enums.MemberScoreDetailReleationStatueEnum;
import cn.iocoder.yudao.module.member.enums.ScoreOperateTypeEnum;
import cn.iocoder.yudao.module.member.enums.ScoreSourceTypeEnum;
import cn.iocoder.yudao.module.member.service.memberUserScore.MemberUserScoreService;
import cn.iocoder.yudao.module.member.service.memberUserScoreDetail.MemberUserScoreDetailService;
import cn.iocoder.yudao.module.member.service.memberUserScoreDetailReleation.MemberUserScoreDetailReleationService;
import cn.iocoder.yudao.module.member.service.memberUserScoreLog.MemberUserScoreLogService;
import cn.iocoder.yudao.module.member.vo.memberUserScoreLog.MemberUserScoreLogCreateReq;
import com.alibaba.excel.util.CollectionUtils;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.stream.Collectors;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.module.infra.enums.ErrorCodeConstants.GET_LOCK_FAILED;
import static cn.iocoder.yudao.module.member.enums.ErrorCodeConstants.*;

@Slf4j
@Service
public class MemberUserScoreApiImpl implements MemberUserScoreApi {

    @Resource
    private MemberUserScoreLogService logService;
    @Resource
    private MemberUserScoreDetailService scoreDetailService;
    @Resource
    private MemberUserScoreService memberUserScoreService;
    @Resource
    private MemberUserScoreDetailReleationService memberUserScoreDetailReleationService;
    @Resource
    private ApplicationContext applicationContext;
    @Resource
    private RedissonClient redissonClient;

    @Override
    @Transactional(rollbackFor = Exception.class)
    public MemberUserScoreOperateRespDTO operateScore(MemberUserScoreOperateReqDTO req) {
        if (req.getMemberId() == null) {
            throw exception(MEMBER_ID_IS_NULL);
        }
        if (req.getScoreCount() <= 0) {
            throw exception(SCORE_COUNT_ERROR);
        }
        String lockKey = "member:operate:score:" + req.getMemberId();
        RLock lock = redissonClient.getLock(lockKey);
        try {
            boolean lockSuccess = lock.tryLock(2, 2, TimeUnit.MINUTES);
            if (!lockSuccess) {
                throw exception(GET_LOCK_FAILED);
            }
            Long logId = saveScoreLog(req);
            List<ScoreDetailChangeDto> detailInfos = saveScoreDetail(req, logId);
            saveScoreDetailReleation(req, detailInfos);
            updateUserScore(req);
            publishDetailExpireEvent(req, detailInfos);
        } catch (InterruptedException e) {
            throw exception(GET_LOCK_FAILED);
        } finally {
            lock.unlock();
        }
        return MemberUserScoreOperateRespDTO.success(req);
    }

    private void publishDetailExpireEvent(MemberUserScoreOperateReqDTO req, List<ScoreDetailChangeDto> detailInfos) {
        if (req.getSourceType().getReverseSource() == null || req.getOperateType() == ScoreOperateTypeEnum.REDUCE) {
            return;
        }
        //判断回退流程，回退的积分是否过期.如果过期再走正常的过期扣积分流程
        MemberScoreDetailExpireEvent memberScoreDetailExpireEvent = new MemberScoreDetailExpireEvent();
        memberScoreDetailExpireEvent.setDetailIds(detailInfos.stream().map(ScoreDetailChangeDto::getDetailId).collect(Collectors.toList()));
        applicationContext.publishEvent(memberScoreDetailExpireEvent);
    }

    private void saveScoreDetailReleation(MemberUserScoreOperateReqDTO req, List<ScoreDetailChangeDto> detailInfos) {
        if (StringUtils.isBlank(req.getReleationId())) {
            return;
        }
        memberUserScoreDetailReleationService.updateReleation(req.getReleationId(), req.getSourceType(), detailInfos);
    }

    @Override
    @Transactional
    public List<MemberUserScoreOperateRespDTO> batchOperateScore(MemberUserScoreBatchOperateReqDTO req) {
        if (CollectionUtils.isEmpty(req.getMemberIds())) {
            throw exception(MEMBER_ID_IS_NULL);
        }
        if (req.getScoreCount() <= 0) {
            throw exception(SCORE_COUNT_ERROR);
        }
        // 校验用户当前积分是否满足扣减要求
        if (req.getOperateType() == ScoreOperateTypeEnum.REDUCE) {
            LambdaQueryWrapper<MemberUserScoreDO> wrapper = Wrappers.lambdaQuery();
            wrapper.in(MemberUserScoreDO::getMemberId, req.getMemberIds());
            List<MemberUserScoreDO> userScoreDOList = memberUserScoreService.list(wrapper);
            List<MemberUserScoreDO> notEnoughScoreList = userScoreDOList.stream()
                    .filter(item -> item.getHoldScore() < req.getScoreCount()).collect(Collectors.toList());
            if (!notEnoughScoreList.isEmpty()) {
                throw exception(MEMBER_SCORE_NOT_ENOUGH);
            }
        }
        return req.getMemberIds().stream().map(memberId -> operateScore(MemberUserScoreOperateReqDTO.builder()
                        .memberId(memberId)
                        .scoreCount(req.getScoreCount())
                        .operateType(req.getOperateType())
                        .sourceType(req.getSourceType())
                        .extParam(req.getExtParam())
                        .build()))
                .collect(Collectors.toList());
    }

    @Override
    public List<ReleationScoreExpireInfoDTO> getScoreExpireInfo(String releationId, ScoreSourceTypeEnum scoreSourceTypeEnum) {
        if (StringUtils.isBlank(releationId)) {
            return Collections.emptyList();
        }
        LambdaQueryWrapper<MemberUserScoreDetailReleationDO> wrapper = Wrappers.lambdaQuery();
        wrapper.eq(MemberUserScoreDetailReleationDO::getReleationId, releationId);
        wrapper.eq(MemberUserScoreDetailReleationDO::getSourceType, scoreSourceTypeEnum.getValue());
        wrapper.eq(MemberUserScoreDetailReleationDO::getStatus, MemberScoreDetailReleationStatueEnum.AVAILABLE.getValue());
        List<MemberUserScoreDetailReleationDO> releationDOList = memberUserScoreDetailReleationService.list(wrapper);
        Set<Long> detailIds = releationDOList.stream().map(MemberUserScoreDetailReleationDO::getDetailId).collect(Collectors.toSet());
        LambdaQueryWrapper<MemberUserScoreDetailDO> detailWrappers = Wrappers.lambdaQuery();
        detailWrappers.in(MemberUserScoreDetailDO::getId, detailIds);
        Map<Long, MemberUserScoreDetailDO> detailIdInfoMap = scoreDetailService.list(detailWrappers).stream()
                .collect(Collectors.toMap(MemberUserScoreDetailDO::getId, Function.identity(), (c1, c2) -> c1));
        return releationDOList.stream().map(releation -> {
            ReleationScoreExpireInfoDTO expireInfoDTO = new ReleationScoreExpireInfoDTO();
            MemberUserScoreDetailDO detailDO = detailIdInfoMap.getOrDefault(releation.getDetailId(), new MemberUserScoreDetailDO());
            expireInfoDTO.setDetailId(detailDO.getId());
            expireInfoDTO.setExpireDate(detailDO.getExpireTime());
            expireInfoDTO.setScoreCount(releation.getScoreCount());
            return expireInfoDTO;
        }).collect(Collectors.toList());
    }

    private void updateUserScore(MemberUserScoreOperateReqDTO req) {
        memberUserScoreService.updateUserScore(MemberUserScoreUpdateReqDTO.builder()
                .memberId(req.getMemberId())
                .scoreCount(req.getScoreCount())
                .operateType(req.getSourceType() == ScoreSourceTypeEnum.MANUAL_OPERATE ?
                        req.getOperateType() : req.getSourceType().getOperateType())
                .sourceType(req.getSourceType())
                .build());
    }

    private List<ScoreDetailChangeDto> saveScoreDetail(MemberUserScoreOperateReqDTO req, Long scoreLogId) {
        return scoreDetailService.updateScoreDetail(MemberUserScoreDetailUpdateReqDto.builder()
                .memberId(req.getMemberId())
                .scoreCount(req.getScoreCount())
                .operateType(req.getSourceType() == ScoreSourceTypeEnum.MANUAL_OPERATE ?
                        req.getOperateType() : req.getSourceType().getOperateType())
                .sourceType(req.getSourceType())
                .scoreLogId(scoreLogId)
                .expireDays(req.getExpireDays())
                .releationId(req.getReleationId())
                .build());
    }

    private Long saveScoreLog(MemberUserScoreOperateReqDTO req) {
        return logService.createScoreLog(MemberUserScoreLogCreateReq.builder()
                .memberId(req.getMemberId())
                .scoreCount(req.getScoreCount())
                .ruleId(req.getRuleId())
                .operateType(req.getSourceType() == ScoreSourceTypeEnum.MANUAL_OPERATE ?
                        req.getOperateType().getValue() : req.getSourceType().getOperateType().getValue())
                .sourceType(req.getSourceType().getValue())
                .extParam(req.getExtParam())
                .build());
    }
}
