package cn.iocoder.yudao.module.member.listener;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.ArrayUtil;
import cn.hutool.json.JSONUtil;
import cn.iocoder.boot.module.order.api.OrderApi;
import cn.iocoder.boot.module.order.api.dto.OrderObjectiveApiDO;
import cn.iocoder.boot.module.order.api.dto.OrderRespDTO;
import cn.iocoder.boot.module.order.enums.OrderStatusApiEnum;
import cn.iocoder.yudao.framework.apollo.core.event.OrderInShippingEvent;
import cn.iocoder.yudao.module.member.api.score.MemberUserScoreApi;
import cn.iocoder.yudao.module.member.api.score.dto.MemberUserScoreOperateReqDTO;
import cn.iocoder.yudao.module.member.dal.dataobject.memberUserScoreLog.MemberUserScoreLogDO;
import cn.iocoder.yudao.module.member.dal.dataobject.scoreRule.ScoreRuleDO;
import cn.iocoder.yudao.module.member.enums.PlatformTypeEnum;
import cn.iocoder.yudao.module.member.enums.ScoreSourceTypeEnum;
import cn.iocoder.yudao.module.member.enums.TransportTypeEnum;
import cn.iocoder.yudao.module.member.enums.YesOrNoTypeEnum;
import cn.iocoder.yudao.module.member.service.memberUserScoreLog.MemberUserScoreLogService;
import cn.iocoder.yudao.module.member.service.scoreRule.ScoreRuleService;
import cn.iocoder.yudao.module.member.vo.scoreRule.extra.ScoreRuleOrderVExtraVO;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.context.event.EventListener;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Objects;

@Component
@AllArgsConstructor
@Slf4j
public class OrderInShippingListener {
    @Resource
    private OrderApi orderApi;
    @Resource
    private ScoreRuleService scoreRuleService;
    @Resource
    private MemberUserScoreApi memberUserScoreApi;
    @Resource
    private MemberUserScoreLogService memberUserScoreLogService;

    @EventListener(OrderInShippingEvent.class)
    public void listen(OrderInShippingEvent event) {
        Long orderId = event.getOrderId();
        String orderNo = event.getOrderNo();
        OrderRespDTO order;
        try {
            order = orderApi.getOrder(orderId, orderNo);
        } catch (Exception e) {
            log.error("Order in shipping listener: Get order error,orderId:{},orderNo:{}", orderId, orderNo);
            return;
        }
        if (Objects.isNull(order)) {
            log.error("Order in shipping listener: The order does not exist,orderId:{},orderNo:{}", orderId, orderNo);
            return;
        }
        // 订单状态应为起运
        if (!Objects.equals(order.getStatus(), OrderStatusApiEnum.IN_SHIPPING.getValue())) {
            return;
        }
        // 判断海运空运
        Integer transportId = order.getTransportId();
        ScoreRuleDO scoreRuleDO = scoreRuleService.getEnabledOrderVScoreRuleByTransportType(transportId);
        if (Objects.isNull(scoreRuleDO)) {
            log.info("Order in shipping listener: No score rule match");
            return;
        }
        // 是否过期，或者活动还没开始
        if (!scoreRuleDO.getEndTime().after(new Date()) || !scoreRuleDO.getStartTime().before(new Date())) {
            log.info("Order in shipping listener: The score rule has expired");
            return;
        }
        // 目的地信息
        OrderObjectiveApiDO orderObjectiveDO = order.getOrderObjectiveDO();
        ScoreRuleOrderVExtraVO extraInfo = JSONUtil.toBean(scoreRuleDO.getExtra(), ScoreRuleOrderVExtraVO.class);
        // 提货点是否包含
        String[] warehouseIds = extraInfo.getReceiveAddr().split(",");
        if (!ArrayUtil.contains(warehouseIds, orderObjectiveDO.getObjectiveWarehouseId().toString())) {
            return;
        }
        // 计算积分
        List<ScoreRuleOrderVExtraVO.OrderVRule> orderVRules = extraInfo.getOrderVRule();
        Integer scoreCount = 0;
        if (transportId == TransportTypeEnum.OCEAN_LCL.getValue()) {
            // 海运算重量
            BigDecimal orgVWeight = order.getOrgVWeight();
            scoreCount = computeScoreCount(orgVWeight, orderVRules);
        } else if (transportId == TransportTypeEnum.SPECIAL_LINE_AIR_FREIGHT.getValue()) {
            // 空运校验渠道
            // 渠道id
            Long channelId = order.getChannelId();
            String[] channels = extraInfo.getChannel().split(",");
            if (!ArrayUtil.contains(channels, channelId.toString())) {
                return;
            }
            // 空运算体积
            BigDecimal orgWVolume = order.getOrgWVolume();
            scoreCount = computeScoreCount(orgWVolume, orderVRules);
        }
        if (scoreCount > scoreRuleDO.getGetScoreOnce()) {
            scoreCount = scoreRuleDO.getGetScoreOnce();
        }

        // 会员id
        Long userId = order.getUserId();
        // 订单入口
        String orderEntryString = extraInfo.getOrderEntry();
        if (StringUtils.isBlank(orderEntryString)) {
            log.info("Order in shipping listener: orderEntry is null");
            return;
        }
        String[] orderEntry = orderEntryString.split(",");
        // TODO：后台下单会员获取待确认
        if (userId <= 0) {
            // 后台下单
            if (!ArrayUtil.contains(orderEntry, String.valueOf(PlatformTypeEnum.BACKEND.getValue()))) {
                log.info("Order in shipping listener: The order is placed by the background, userId:{}", userId);
                return;
            }
            //userId = order.getCustomerId();
        } else if (!ArrayUtil.contains(orderEntry, String.valueOf(PlatformTypeEnum.APP.getValue())) && !ArrayUtil.contains(orderEntry, String.valueOf(PlatformTypeEnum.WEB.getValue()))) {
            log.info("Order in shipping listener: The order is placed by the app or web, userId:{}", userId);
            return;
        }
        // 是否首单，首单只加一次
        if (extraInfo.getFirstOrder() == YesOrNoTypeEnum.YES.ordinal()) {
            if (!orderApi.isFirstOrder(userId)) {
                log.info("Order in shipping listener: Not first order");
                return;
            }
        }
        // 校验累计最高分
        Integer userScoreTotalCount = 0;
        LambdaQueryWrapper<MemberUserScoreLogDO> scoreLogQueryWrapper = new LambdaQueryWrapper<>();
        scoreLogQueryWrapper.eq(MemberUserScoreLogDO::getMemberId, userId)
                .eq(MemberUserScoreLogDO::getRuleId, scoreRuleDO.getId());
        List<MemberUserScoreLogDO> memberUserScoreLogs = memberUserScoreLogService.selectList(scoreLogQueryWrapper);
        if (CollUtil.isNotEmpty(memberUserScoreLogs)) {
            for (MemberUserScoreLogDO memberUserScoreLog : memberUserScoreLogs) {
                userScoreTotalCount += memberUserScoreLog.getScoreCount();
            }
        }
        if (userScoreTotalCount >= scoreRuleDO.getMaxScoreTotal()) {
            log.info("Order in shipping listener: The user has reached the maximum score,userId:{},scoreRuleId:{}", userId, scoreRuleDO.getId());
            return;
        }
        //if (userScoreTotalCount + scoreCount > scoreRuleDO.getMaxScoreTotal()) {
        //    scoreCount = scoreRuleDO.getMaxScoreTotal() - userScoreTotalCount;
        //}
        // 增加积分
        try {
            HashMap<String, Object> map = new HashMap<>();
            map.put("orderId", orderId);
            memberUserScoreApi.operateScore(MemberUserScoreOperateReqDTO.builder()
                    .memberId(userId)
                    .scoreCount(scoreCount)
                    .sourceType(ScoreSourceTypeEnum.ORDER_V)
                    .ruleId(scoreRuleDO.getId())
                    .expireDays(scoreRuleDO.getScorePeriod())
                    .extParam(map)
                    .build()
            );
        } catch (Exception e) {
            log.error("Order in shipping listener: operateScore error", e);
        }
    }

    /**
     * 根据规则计算获取积分
     *
     * @param weightOrVolume 重量或体积
     * @param orderVRules    规则
     * @return
     */
    private Integer computeScoreCount(BigDecimal weightOrVolume, List<ScoreRuleOrderVExtraVO.OrderVRule> orderVRules) {
        Integer scoreCount = 0;
        int w = weightOrVolume.setScale(0, RoundingMode.HALF_UP).intValue();
        for (int i = orderVRules.size() - 1; i >= 0; i--) {
            ScoreRuleOrderVExtraVO.OrderVRule orderVRule = orderVRules.get(i);
            if (w > orderVRule.getHigh()) {
                scoreCount = orderVRule.getScore();
                break;
            }
            if (orderVRule.getLow() <= w && orderVRule.getHigh() > w) {
                scoreCount = orderVRule.getScore();
                break;
            }
        }
        return scoreCount;
    }
}
